import os, osproc, streams, strtabs

type HandleProcMsgCB* = proc (cmd: string, outMsg: TaintedString, errMsg: TaintedString): void

proc raiseEx*(reason: string): void =
  raise newException(Exception, reason)

proc envToTable*(): StringTableRef =
  result = newStringTable()

  for k, v in envPairs():
    result[k] = v

proc waitForWithOutput*(p: Process, msgCB: HandleProcMsgCB,
                       procCmd: string = ""):
      tuple[output: TaintedString, error: TaintedString, exitCode: int] =

  var pout = outputStream(p)
  var perr = errorStream(p)

  result = (TaintedString"", TaintedString"", -1)
  var line = newStringOfCap(120).TaintedString
  while true:
    if pout.readLine(line):
      if msgCB != nil: msgCB(procCmd, line, nil)
      result[0].string.add(line.string)
      result[0].string.add("\n")
    elif perr.readLine(line):
      if msgCB != nil: msgCB(procCmd, nil, line)
      result[1].string.add(line.string)
      result[1].string.add("\n")
    else:
      result[2] = peekExitCode(p)
      if result[2] != -1: break
  close(p)
  
proc exec*(command: string, workingDir: string = "",
          args: openArray[string] = [], env: StringTableRef = nil,
          options: set[ProcessOption] = {poUsePath},
          msgCB: HandleProcMsgCB = nil):
            tuple[output: TaintedString, error: TaintedString, exitCode: int]
            {.tags: [ExecIOEffect, ReadIOEffect], gcsafe.} =

  var p = startProcess(command, workingDir, args, env, options)
  result = waitForWithOutput(p, msgCb, command)

proc loadEnv*(): StringTableRef =
  result = newStringTable()

  for k, v in envPairs():
    result[k] = v
